﻿/*
 * Created by SharpDevelop.
 * User: phil
 * Date: 2010.09.17.
 * Time: 11:19
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using System;
using System.Collections.Generic;
using System.Xml;
//
using TetheredSun.Mathematics;

namespace TetheredSun.SignalAnalysis
{
	/// <summary>
	/// Description of ScaledSignal.
	/// </summary>
	public class ScaledSignal : List<double>, IXmlConvertible, ISignal // TODO: Consider a better implementation instead of extending List<T>.
	{
		protected IScale xScale;


		public ScaledSignal(IScale xScale, IList<double> values) : base(values)
		{
			this.xScale = xScale;
		}

		public ScaledSignal(IScale xScale) : base(xScale.Count)
		{
			this.xScale = xScale;
		}

		public ScaledSignal() : this(null)
		{
		}

		public IList<double> X {
			get { return xScale.Values; }
		}
		
		public IList<double> Y {
			get { return this; }
		}

		public double Mean {
			get { return this.GetMean(); }
		}
		
		public bool IsReadOnly {
			get { return (this as IList<double>).IsReadOnly; }
		}
		
		public new double this[int index] {
			get { return base[index]; }
			set {
				if (index >= this.Count) {
					base.Add(value);
				} else {
					base[index] = value;
				}
			}
		}
		
		public IScale XScale {
			get { return xScale; }
			set { xScale = value; }
		}


		public ISignal Reciprocal()
		{
			ScaledSignal result = new ScaledSignal(xScale.Clone());

			for (int i = 0; i < this.Count; i++) {
				result[i] = 1.0 / this[i];
			}

			return result;
		}

		public ISignal SubtractMean()
		{
			return (this -this.GetMean());
		}

		
		public new void Add(double item)
		{
			xScale.Count++;
			base.Add(item);
		}
		
		public new void AddRange(IEnumerable<double> collection)
		{
			xScale.Count += (collection as ICollection<double>).Count;
			
			base.AddRange(collection);
		}
		
		public new void Clear()
		{
			xScale = null;
			base.Clear();
		}
		
		public new void Insert(int index, double item)
		{
			xScale.Count++;
			base.Insert(index, item);
		}
		
		public new void InsertRange(int index, IEnumerable<double> collection)
		{
			xScale.Count += (collection as ICollection<double>).Count;
			
			base.InsertRange(index, collection);
		}
		
		public new bool Remove(double item)
		{
			if (base.Remove(item)) {
				xScale.Count--;
				return true;
			} else {
				return false;
			}
		}
		
		public new int RemoveAll(Predicate<double> match)
		{
			int matches = base.RemoveAll(match);
			
			xScale.Count -= matches;
			
			return matches;
		}
		
		
		public new void RemoveAt(int index)
		{
			base.RemoveAt(index);
			xScale.Count--;
		}
		
		public new void RemoveRange(int index, int count)
		{
			base.RemoveRange(index, count);
			xScale.Count -= count;
		}
		
		public void FromXml(XmlElement xmlElement)
		{
			if (xmlElement.Name != "ScaledSignal")
				throw new ArgumentException("Cannot load ScaledSignal data from a different xml element.");
			if (xmlElement.HasChildNodes) {
				foreach (XmlNode child in xmlElement.ChildNodes) {
					if (!(child is XmlElement))
						continue;
					switch (child.Name) {
						case "Scale":
							if (child.Attributes["type"].Value == "ByBounds") {
								xScale = new ScaleByBounds();
							} else {
								xScale = new ScaleByIncrement();
							}
							xScale.FromXml(child as XmlElement);
							break;
						case "Data":
							this.Clear();
							this.AddRange(ConversionTools.GetDoubleList(Convert.FromBase64String(child.InnerXml)));
							break;
						default:
							break;
					}
				}
			}
		}

		public XmlElement ToXml(XmlDocument context)
		{
			XmlElement xml = context.CreateElement("ScaledSignal");
			XmlElement child;

			if (xScale == null)
				return xml;

			child = xScale.ToXml(context);
			xml.AppendChild(child);

			child = context.CreateElement("Data");
			child.SetAttribute("count", this.Count.ToString());
			child.SetAttribute("binaryEncoding", "Base64");
			child.InnerXml = Convert.ToBase64String(ConversionTools.GetBytes(this), Base64FormattingOptions.InsertLineBreaks);
			xml.AppendChild(child);

			return xml;
		}


		public static ScaledSignal Average(IList<ScaledSignal> scaledSignals)
		{
			if (scaledSignals == null)
				throw new NullReferenceException();
			if (scaledSignals.Count < 1)
				throw new InvalidOperationException("The list of signals is empty.");
			ScaledSignal average = new ScaledSignal(scaledSignals[0].XScale.Clone());

			for (int i = 0; i < scaledSignals.Count; i++) {
				for (int j = 0; (j < average.Count) && (j < scaledSignals[i].Count); j++) {
					average[j] += scaledSignals[i][j];
				}
			}

			return average / scaledSignals.Count;
		}

		public static ScaledSignal Normalise(ScaledSignal scaledSignal)
		{
			double min = Double.MaxValue;
			double max = Double.MinValue;
			double range;

			for (int i = 0; i < scaledSignal.Count; i++) {
				if (scaledSignal[i] < min)
					min = scaledSignal[i];
				if (scaledSignal[i] > max)
					max = scaledSignal[i];
			}

			range = max - min;
			return (1.0 / range) * (scaledSignal - min);
		}


		public static ScaledSignal operator -(ScaledSignal scaledSignal)
		{
			ScaledSignal result = new ScaledSignal(scaledSignal.xScale.Clone());

			for (int i = 0; i < scaledSignal.Count; i++) {
				result[i] = -scaledSignal[i];
			}

			return result;
		}

		public static ScaledSignal operator +(ScaledSignal scaledSignal, double d)
		{
			ScaledSignal result = new ScaledSignal(scaledSignal.xScale.Clone());

			for (int i = 0; i < scaledSignal.Count; i++) {
				result[i] = scaledSignal[i] + d;
			}

			return result;
		}

		public static ScaledSignal operator +(double d, ScaledSignal scaledSignal)
		{
			return scaledSignal + d;
		}

		public static ScaledSignal operator -(ScaledSignal scaledSignal, double d)
		{
			return scaledSignal + (-d);
		}

		public static ScaledSignal operator -(double d, ScaledSignal scaledSignal)
		{
			return d + (-scaledSignal);
		}

		public static ScaledSignal operator *(double d, ScaledSignal scaledSignal)
		{
			ScaledSignal result = new ScaledSignal(scaledSignal.xScale.Clone());

			for (int i = 0; i < scaledSignal.Count; i++) {
				result[i] = scaledSignal[i] * d;
			}

			return result;
		}

		public static ScaledSignal operator *(ScaledSignal scaledSignal, double d)
		{
			return d * scaledSignal;
		}

		public static ScaledSignal operator /(ScaledSignal scaledSignal, double d)
		{
			ScaledSignal result = new ScaledSignal(scaledSignal.xScale.Clone());

			for (int i = 0; i < scaledSignal.Count; i++) {
				result[i] = scaledSignal[i] / d;
			}

			return result;
		}

		public static ScaledSignal operator /(double d, ScaledSignal scaledSignal)
		{
			return d * (scaledSignal.Reciprocal() as ScaledSignal);
		}

		// TODO: Implement operations between signals.
	}
}
