﻿/*
 * A wrapper to some FFTW methods (http://www.fftw.org/).
 * Based on Tamás Szalay's wrapper classes (http://www.sdss.jhu.edu/~tamas/bytes/fftwcsharp.html).
 * 
 */
using System;
using System.Runtime.InteropServices;
//
using TetheredSun.Mathematics;
using TetheredSun.Spectral;

namespace TetheredSun.Spectral.Fftw
{
	/// <summary>
	/// A wrapper around FFTW.
	/// </summary>
	public static class Fourier
	{
		public static Complex[] ComplexTransformBackwards(Complex[] samples)
		{
			int length = samples.Length;
            double[] inputOutput = new double[2 * length];
            Complex[] transform = new Complex[length];
            GCHandle inputOutputHandle = GCHandle.Alloc(inputOutput, GCHandleType.Pinned);
            IntPtr plan;
            
            for (int i = 0; i < length; i++) {
            	inputOutput[2 * i] = samples[i].Real;
            	inputOutput[2 * i + 1] = samples[i].Imaginary;
            }
            
            plan = Interop.OneDimensionalComplexToComplex(length, inputOutputHandle.AddrOfPinnedObject(), inputOutputHandle.AddrOfPinnedObject(), FftwDirection.Backwards, FftwFlags.Estimate);
            Interop.Execute(plan);
            
            for (int i = 0; i < length; i++) {
            	transform[i] = new Complex(inputOutput[2 * i] / length, inputOutput[2 * i + 1] / length);
            }
            
            Interop.DestroyPlan(plan);
            Interop.Cleanup();
            inputOutputHandle.Free();
            
            return transform;
		}
		
		public static Complex[] ComplexTransformForwards(Complex[] samples)
		{
			int length = samples.Length;
            double[] inputOutput = new double[2 * length];
            Complex[] transform = new Complex[length];
            GCHandle inputOutpuHandle = GCHandle.Alloc(inputOutput, GCHandleType.Pinned);
            IntPtr plan;
            
            for (int i = 0; i < length; i++) {
            	inputOutput[2 * i] = samples[i].Real;
            	inputOutput[2 * i + 1] = samples[i].Imaginary;
            }
            
            plan = Interop.OneDimensionalComplexToComplex(length, inputOutpuHandle.AddrOfPinnedObject(), inputOutpuHandle.AddrOfPinnedObject(), FftwDirection.Forwards, FftwFlags.Estimate);
            Interop.Execute(plan);
            
            for (int i = 0; i < length; i++) {
            	transform[i] = new Complex(inputOutput[2 * i], inputOutput[2 * i + 1]);
            }
            
            Interop.DestroyPlan(plan);
            Interop.Cleanup();
            inputOutpuHandle.Free();
            
            return transform;
		}
		
		private static Complex[] CrossSpectrum(double[] signal1, double[] signal2, SpectrumType type, double factor)
		{
			int length = Math.Min(signal1.Length, signal2.Length);
			int spectrumLength = (type == SpectrumType.OneSided) ? length / 2 : length;
			Complex complex;
			Complex[] transform1 = RealTransformForwards(signal1);
			Complex[] transform2 = RealTransformForwards(signal2);
			Complex[] crossSpectrum = new Complex[spectrumLength];
			
			if (type == SpectrumType.OneSided) {
				crossSpectrum[0] = transform1[0] * transform2[0] * factor;
				
				for (int i = 1; i < spectrumLength; i++) {
					complex = Complex.Conjugate(transform1[i]) * transform2[i];
					crossSpectrum[i] = new Complex(2.0 * factor * complex.Real, 2.0 * factor * complex.Imaginary);
				}
			} else {
				for (int i = 0; i < length; i++) {
					complex = Complex.Conjugate(transform1[i]) * transform2[i];
					crossSpectrum[i] = new Complex(complex.Real * factor, complex.Imaginary * factor);
				}
			}
			
			return crossSpectrum;
		}
		
		public static Complex[] CrossSpectrum(double[] signal1, double[] signal2, SpectrumType type)
		{
			int length = Math.Min(signal1.Length, signal2.Length);
			double factor = 1.0 / (length * length);
			
			return CrossSpectrum(signal1, signal2, type, factor);
		}
		
		public static Complex[] CrossSpectralDensity(double[] signal1, double[] signal2, SpectrumType type, double dt, out double df)
		{
			int length = Math.Min(signal1.Length, signal2.Length);
			double factor = dt / length;
			
			df = 1.0 / (length * dt);
			
			return CrossSpectrum(signal1, signal2, type, factor);
		}
		
		public static double[] PowerSpectralDensity(double[] samples, SpectrumType type, double dt, out double df)
		{
			df = 1.0 / (samples.Length * dt);
			
			// S(f) ~ |X(f)|^2 / T
			// The Fourier transform and the discrete Fourier transform at the same frequency differ in a factor of dt:
			// X(n * df) = DFT_n * dt.
			// S(f) ~ |DFT * dt|^2 / (N * dt) = (|DFT|^2 * dt) / N = |DFT|^2 / (N / dt).
			return PowerSpectrum(samples, type, dt / samples.Length);
		}
		
		public static double[] PowerSpectrum(double[] samples, SpectrumType type)
		{ 
			return PowerSpectrum(samples, type, 1.0 / (samples.Length * samples.Length));
		}
		
		private static double[] PowerSpectrum(double[] samples, SpectrumType type, double factor)
		{
			int length = samples.Length;
			int spectrumLength = (type == SpectrumType.OneSided) ? length / 2 : length;
			double abs1;
			double abs2;
			Complex[] transform = RealTransformForwards(samples);
			double[] spectrum = new double[spectrumLength];
			
			if (type == SpectrumType.OneSided) {
				abs1 = Complex.Abs(transform[0]);
				spectrum[0] = abs1 * abs1 * factor;
				
				for (int i = 1; i < spectrumLength; i++) {
					abs1 = Complex.Abs(transform[i]);
					abs2 = Complex.Abs(transform[length - i]);
					spectrum[i] = (abs1 * abs1 + abs2 * abs2) * factor;
				}
			} else {
				for (int i = 0; i < length; i++) {
					abs1 = Complex.Abs(transform[i]);
					spectrum[i] = abs1 * abs1 * factor;
				}
			}
			
			return spectrum;
		}
		
		// The complex sequence is assumed to contain the redundant elements as well, so it is of the same length as the real result. 
		public static double[] RealTransformBackwards(Complex[] samples)
		{
			int length = samples.Length;
			int usefulLength = length / 2 + 1;
			double[] inputOutput = new double[2 * length];
			double[] transform = new double[length];
			GCHandle inputOutputHandle = GCHandle.Alloc(inputOutput, GCHandleType.Pinned);

            IntPtr plan;
            
            for (int i = 0; i < usefulLength; i++) {
            	inputOutput[2 * i] = samples[i].Real;
            	inputOutput[2 * i + 1] = samples[i].Imaginary;
            }
            
            plan = Interop.OneDimensionalComplexToReal(length, inputOutputHandle.AddrOfPinnedObject(), inputOutputHandle.AddrOfPinnedObject(), FftwFlags.Estimate);
            Interop.Execute(plan);
            
            // Perform scaling.
            for (int i = 0; i < length; i++) {
            	transform[i] = inputOutput[i] / length;
            }
            
            Interop.DestroyPlan(plan);
            Interop.Cleanup();
            inputOutputHandle.Free();
            
            return transform;
		}
		
		public static Complex[] RealTransformForwards(double[] samples)
		{
            int length = samples.Length;
            int nyquistLength = length / 2;
            double[] inputOutput = new double[length + 2];	// Output consists of n/2 + 1 complex numbers, requires a double array 2 * (n/2 + 1) long.
            Complex current;
            Complex[] transform = new Complex[length];
            GCHandle inputOutputHandle = GCHandle.Alloc(inputOutput, GCHandleType.Pinned);
            IntPtr plan;
            
            Array.Copy(samples, inputOutput, length);
            plan = Interop.OneDimensionalRealToComplex(length, inputOutputHandle.AddrOfPinnedObject(), inputOutputHandle.AddrOfPinnedObject(), FftwFlags.Estimate);
            Interop.Execute(plan);
            
            transform[0] = new Complex(inputOutput[0], inputOutput[1]);
            
            for (int i = 1; i < nyquistLength; i++) {
            	current = new Complex(inputOutput[2 * i], inputOutput[2 * i + 1]);
            	transform[i] = current;
            	transform[length - i] = Complex.Conjugate(current);
            }
            
            transform[nyquistLength] = new Complex(inputOutput[2 * nyquistLength], inputOutput[2 * nyquistLength + 1]);
            
             if ((length % 2) != 0) {
            	transform[nyquistLength + 1] = Complex.Conjugate(transform[nyquistLength]);
            }
            
            Interop.DestroyPlan(plan);
            Interop.Cleanup();
            inputOutputHandle.Free();
            
            return transform;
		}
	}
}
